@* Curl Socket Example.

Sorry I write this in CWEB since its much easier to provide one file, while
testing the dependencies of several libraries in one sourcecode file, which gets
split during processing.

@ So lets start with the entry point and how this example is deployed.
@c
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <curl/curl.h>
#include "qt-part.h"
@<global operation data@>@;
@<global functions@>@;
@<curl callbacks@>@;
@<curly functions@>@;
int main( int argc, char** argv )
{
	int io_channel_counter = 0; /* when no io channels are open anymore we
				       can quit the application */
	struct global_data global;
	struct test_read_data* trd = calloc(1,sizeof(struct test_read_data));
	int code;

	curl_global_init(CURL_GLOBAL_ALL);
	create_test_app();

	trd->io_channel_counter = &io_channel_counter, io_channel_counter++;
	create_io_channel( 0, 0, trd, test_read_cb ); /* register stdin */

	global.curl = curl_multi_init();
	global.io_channel_counter = &io_channel_counter; io_channel_counter++;
	setup_curl_multi_urls( &global, "http://curl.haxx.se/download" );

	while ( CURLM_CALL_MULTI_PERFORM == 
		(code = curl_multi_socket_all(global.curl, &global.handles)) );

	exec_test_app();
	if ( trd->some_fd_context ) free( trd->some_fd_context );
	curl_multi_cleanup( global.curl );
	return 0;
}

@ Global operation data
@<global operation data@>=
struct global_data {
	CURLM* curl;
	int handles;
	int* io_channel_counter;
};

@* Curl Functions.

@ Setup a single URL.
@c
void* setup_curl_url( struct global_data* global, const char* url )
{
	CURL* curl = curl_easy_init();
	struct url_opdat* url_dat = calloc( 1, sizeof(struct url_opdat) );
	url_dat->global = global;
	url_dat->url = malloc(strlen(url)+1);
	strcpy( url_dat->url, url );
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION,  incoming_list );
	curl_easy_setopt(curl, CURLOPT_HEADERFUNCTION,  incoming_list );
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, url_dat );
	curl_easy_setopt(curl, CURLOPT_HEADERDATA, url_dat );
	curl_easy_setopt(curl, CURLOPT_NOBODY, 1); /* set a head request */
	curl_easy_setopt(curl, CURLOPT_URL, url );
	return curl;

}

@ Setup a line of urls.
@c
void setup_curl_multi_urls( struct global_data* global, const char* base_url )
{
	char url_buf[1024];
	unsigned i;
	CURL* curl;

	for ( i = 0; i< 25; ++i ) {
		sprintf( url_buf, "%s/curl-7.%d.0.tar.bz2", base_url, i );
		curl = setup_curl_url( global, url_buf );
		if ( curl != NULL ) {
			curl_multi_add_handle(global->curl, curl);
		}
	}
	curl_multi_setopt(global->curl, CURLMOPT_SOCKETFUNCTION, socket_cb );
	curl_multi_setopt(global->curl, CURLMOPT_SOCKETDATA, global );
}

@* Curl Callbacks.
@<curl callbacks@>=
static void socket_cb( CURL* c, curl_socket_t fd, int action,
		void* inf, void* priv );
static void io_socket_cb( void* iof, void* cb_vp, int fd );
static size_t incoming_list( void* dat, size_t len, size_t n, void* req_vp );


@ Curl Socket Callback.
@d DEBUG_HANDLES 1
@c
void socket_cb( CURL* c, curl_socket_t fd, int action,
		void* inf, void* priv )
{
	struct socket_opdat* dat = priv;
	struct global_data* global = inf;
#if DEBUG_HANDLES
	fprintf(stdout, "CURL SOCKET CB %d", fd );
#endif
	if ( dat == NULL ) {
		dat = calloc(1, sizeof(struct socket_opdat));
		dat->global = global;
		curl_multi_assign( global->curl, fd, dat );
	}
	switch ( action ) {
		case CURL_POLL_NONE:
#if DEBUG_HANDLES
			fprintf(stdout, " NONE\n" );
#endif
			break;
		case CURL_POLL_IN: 
#if DEBUG_HANDLES
			fprintf( stdout, " IN %p\n", dat );
#endif
			if ( dat->out_iof ) destroy_io_channel(&dat->out_iof);
			if ( !dat->in_iof )
				dat->in_iof = create_io_channel( fd, 0,
						dat, io_socket_cb );
			break;
		case CURL_POLL_OUT: 
#if DEBUG_HANDLES
			fprintf( stdout, " OUT %p\n", dat );
#endif
			if ( dat->in_iof ) destroy_io_channel(&dat->in_iof);
			if ( !dat->out_iof )
				dat->out_iof = create_io_channel( fd, 1,
						dat, io_socket_cb );
			break;
		case CURL_POLL_INOUT: 
#if DEBUG_HANDLES
			fprintf( stdout, " INOUT %p\n", dat );
#endif
			if ( !dat->out_iof ) 
				dat->out_iof = create_io_channel( fd, 1, dat,
						io_socket_cb );
			if ( !dat->in_iof )
				dat->in_iof = create_io_channel( fd, 0, dat,
						io_socket_cb );
			break;
		case CURL_POLL_REMOVE:
#if DEBUG_HANDLES
			fprintf( stdout, " REMOVE %p\n", dat );
#endif
			if ( dat->out_iof ) destroy_io_channel(&dat->out_iof);
			if ( dat->in_iof ) destroy_io_channel(&dat->in_iof);
			if ( dat != NULL ) free( dat ); /* FIXME DEBUG FIELDS */
			break;
		default:
#if DEBUG_HANDLES
			fprintf( stdout, " unkown\n");
#endif
			break;
	}
}

@ Curl IO Callback.
@c
void io_socket_cb( void* iof, void* dat_vp, int fd )
{
	struct socket_opdat* dat = dat_vp;
	struct global_data* global = dat->global;
	int code;
	while ( CURLM_CALL_MULTI_PERFORM == 
		(code = curl_multi_socket_action(global->curl,
					 fd, 0, &global->handles)) );
	if ( global->handles == 0 ) {
		close_io_channel( global->io_channel_counter );
	}
}

@ Curl incoming callback.
@c
size_t incoming_list( void* dat, size_t len, size_t n, void* req_vp )
{
	struct url_opdat* url_dat = req_vp;
	char* buf;
	buf = malloc ( len*n +1 );
	memcpy( buf, dat, len*n );
	buf[len*n] = '\0';
	if ( len*n < 5 ) { free(buf); return len*n; }
	if ( memcmp( buf, "HTTP/", 5 ) == 0 ) {
		unsigned p = 6;
		long http_code;
		int ok = 0;
		char* e;
		while ( buf[p] != ' '  && p<len*n ) ++p;
		if ( buf[p] == ' ' ) {
			e = buf+p;
			http_code = strtol(buf+p, &e, 10);
			if ( e > buf+p && e-buf < len*n ) {
				if ( e[0] == ' ' ) ok = 1;
			}
		}
		if ( ok ) {
			fprintf( stdout, "INCOMING[%s]:%ld\n", 
					url_dat->url, http_code );
		}

	}
	free(buf);
	return len*n;
}

@ Closing any channel.
@<global functions@>=
void close_io_channel( int* io_channel_counter );

@ Global channel close counter.
@c
void close_io_channel( int* io_channel_counter ) {
	if ( *io_channel_counter > 1 )
		(*io_channel_counter)--;
	else
		call_test_app_quit();
}


@ Each url gets some operation data.
@<curl multi operation data@>=
struct url_opdat {
	struct global_data* global;
	char* url;
};
struct socket_opdat {
	struct global_data* global;
	void* out_iof, *in_iof;
};

@ Two functions create the complete curl setup.
@<curl multi setup@>=
extern void setup_curl_multi_urls( struct global_data* global,
		const char* base_url );
extern void* setup_curl_url( struct global_data* global, const char* url );



@* Simple Functions. A simple read function.
@c
void test_read_cb( void* iof, void* dat, int fd )
{
	char buf[1024];
	struct test_read_data* trd = dat;
	int nread = 0;
	nread = read( fd, buf, 1024 );
	if ( nread <= 0 ) {
		fprintf( stdout, "close on %d\n", fd );
		destroy_io_channel( &iof );
		close_io_channel( trd->io_channel_counter );
	}
	else
		fprintf( stdout, "read %d bytes from %d\n", nread, fd );
	trd->status = 0;
}

@ The curl specific functions, data types and callbacks
@<curly functions@>=

@<curl multi operation data@>@;
@<curl multi setup@>@;

@<test read record@>@;
@<test read cb@>@;

@ The Test read simply reads out a socket or file descriptor.
@<test read record@>=
struct test_read_data {
	char* some_fd_context;
	int status;
	int* io_channel_counter;
};

@
@<test read cb@>=
extern void test_read_cb( void* iof, void* dat, int fd );

@ Export an Interface for C.
@s ProgramIOSelect int
@s QSocketNotifier int
@s QCoreApplication int
@s QObject int
@s call_io_function int
@s Type int
@s QSocketNotifier int
@(qt-part.h@>=@q)@>
#ifndef _QT_PART_H_
#define _QT_PART_H_ 1
#if defined __cplusplus
extern "C" {
#endif

@h
@<qt application prototypes@>@;
@<qt io prototypes@>@;

#if defined __cplusplus
}
#endif
#endif


@* Qt Part of the test.  The Qt Part of this test consists of a simple
QApp creation and the IO mechanism.
@(qt-part.cc@>=@q)@>
@<qt application code@>@;
@<qt io events@>@;
#include "moc_qt-part.cc"

@* Asynchronous Operation:IO.


@ Prototypes for io to export some functions to "C".
@<qt io prototypes@>=
typedef void (*call_io_function)( void* iof, void* dat, int fd );
extern void* create_io_channel( int fd, int type, void* dat,
		call_io_function f );
extern void destroy_io_channel( void** X );
extern void enable_io_channel( void* ios_ptr, int enable );


@ IO Event Class specification.
@<qt io events@>=
#include <QtCore/QSocketNotifier>

class ProgramIOSelect: public QSocketNotifier
{ Q_OBJECT@[@]@/
public:
	ProgramIOSelect( QObject* parent,
			void* dat, call_io_function ciof, int fd,
			unsigned type ):
		QSocketNotifier( fd, (QSocketNotifier::Type)type, parent ),
		m_data(dat),
		m_io_function(ciof)
	{
		connect( this, SIGNAL(activated(int)),
			this, SLOT(call_io(int)) );
		setEnabled(true);
	}

public slots:
	void call_io(int fd) { m_io_function( this, m_data, fd ); }

private:
	void* m_data;
	call_io_function m_io_function;
};


void* create_io_channel( int fd, int type, void* dat, call_io_function f )
{
	ProgramIOSelect* new_io= new ProgramIOSelect( NULL, dat, f, fd, type );
	return new_io;
}

void destroy_io_channel( void** X )
{
	ProgramIOSelect* del_io = (ProgramIOSelect*)(*X);
	del_io->setEnabled(false);
	delete del_io;
	*X=NULL;
}

void enable_io_channel( void* ios_ptr, int enable )
{
	ProgramIOSelect* io = (ProgramIOSelect*)(ios_ptr);
	io->setEnabled(enable);
}


@* Qt Application Part.
@<qt application prototypes@>=
extern void* create_test_app();
extern void call_test_app_quit();
extern void exec_test_app();

@
@d TApp ((TestApp*)get_test_app())
@<qt application code@>=
#include <QtCore/QCoreApplication>
#include "qt-part.h"

static char* _m_argv[2] = {"it's me", NULL};
static int _m_argc = 1;
class TestApp: public QCoreApplication { Q_OBJECT
public:
	TestApp(): QCoreApplication(_m_argc, _m_argv) {}
	~TestApp() {}
};


void* create_test_app() { return new TestApp(); }
void* get_test_app() { return QCoreApplication::instance(); }
void call_test_app_quit() { TApp->quit(); }
void exec_test_app() { TApp->exec(); }

@* Index.
